#!/usr/bin/env python

'''
Web Log Url Parsing class.

Contents:
- Parser: logfile url parsing class
  p_log = weblog.url.Parser(log object)
  methods:
	- p_log.getlogent()
  variables:
	- p_log.cache_size  [ maximum size of url cache ]
  attributes:
    - url_scheme      /  ref_scheme - (http|ftp|gopher...)
    - url_host        /  ref_host 
    - url_path        /  ref_path
    - url_parameters  /  ref_parameters - (section after ';')
    - url_query       /  ref_query - (section after '?')
    - url_fragment    /  ref_fragment - section after '#')
    - all attributes of the log object are available as well.

  This class will parse the url and referer (if available) into their
  respective components. It will also replace each with the unparsed
  result of those components; this assures that the input is fully 
  conformant and in sync with the components.

- test: test function
'''


# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.


__version__ = '0.99'


from urlparse import urlparse, urlunparse
from string import lower

class Parser:
	def __init__(self, log):
		self.log = log
		self.cache_size = 10000
		self._cache = {'url': {}, 'ref': {}}
		self._referer_present = 0
		if hasattr(self.log, 'referer'):
			self._referer_present = 1


	def __getattr__(self, attr):
		try:
			return getattr(self.log, attr)
		except AttributeError:
			raise AttributeError, attr


	def getlogent(self):
		''' Increment position in the log and populate requested attributes '''

		if self.log.getlogent():
			### parse url
			if not self._cache['url'].has_key(self.log.url):
				self._cache['url'][self.log.url] = self._parse(self.log.url, 'url')
			(	self.url_scheme, 
				self.url_host, 
				self.url_path, 
				self.url_parameters,
				self.url_query, 
				self.url_fragment, 
				self.url) = self._cache['url'][self.log.url]

			### parse referer
			if self._referer_present:
				if not self._cache['ref'].has_key(self.log.referer):
					self._cache['ref'][self.log.referer] = self._parse(self.log.referer, 'ref')
				(	self.ref_scheme, 
					self.ref_host, 
					self.ref_path, 
					self.ref_parameters,
					self.ref_query, 
					self.ref_fragment, 
					self.referer) = self._cache['ref'][self.log.referer]
			return 1
		else:
			return 0


	def _parse(self, url, url_type):
		if len(self._cache[url_type]) > self.cache_size:
			self._cache[url_type] = {}
		p = urlparse(url)		
		parsed = (p[0], lower(p[1])) + p[2:]
		return parsed + (urlunparse(parsed),)


def test():
	''' basic test suite- modify at will to test full functionality '''

	import sys
	from weblog import combined

	file = sys.stdin
	log = combined.Parser(file)

	p_log = Parser(log)			# url parser

	while p_log.getlogent():
		print "%s\n%s %s %s" % (log.url, p_log.url, p_log.url_path, p_log.url_query)
		print "%s\n%s %s %s" % (log.referer, p_log.referer, p_log.ref_host, p_log.ref_path)
		print



if __name__ == '__main__':
	test()




# This class is messy and undocumented; will revisit later. 


class Accountant:
	''' keep statistics in a easy-to-use format '''

	def __init__(self):
		self._func = {}		# stores functions that determine stats
		self._count = {}	# stores the stats
	
	def __setattr__(self, name, attr):
		if name[0] != '_': 
			self._func[name] = attr
		else:
			self.__dict__[name] = attr
	
	def __getattr__(self, attr):
		if self._count.has_key(attr):
			if self._count[attr] != 0:
				return self._count[attr]
		return 0
		
	def enter(self, item, ref):
		''' enter data and process according to set rules '''
	
		for name, func in self._func.items(): 
			fullname = ref + '_' + name
			self._count[fullname] = self._count.get(fullname, 0L) + func(item) 
			



#!/usr/bin/env python

'''
Web Log Limiting classes.

Contents:

- Path: directory/page limiting class
  pl_log = weblog.limit.Path(log object)
  methods:
    - pl_log.getlogent()
  variables:
    - pl_log.page_limit [ list of page names/file types to limit to ]
    - pl_log.page_exclude [ list of page names/file types to exclude ]
    - pl_log.path_limit [ list of specific pages/directories to limit to ]
    - pl_log.path_exclude [ list of specific pages/directories to exclude ]
    - pl_log.cache_size (size of page cache)
  read-only variables:
    - pl_log.num_skipped (number of lines that have been passed over)
  attributes:
    - all attributes from the log object are available.

  Page/Directory limiting requires previous use of weblog.url.Parse. Page
  and directory limitations are case sensitive.
  
  The page_limit and page_exclude specifications are for excluding a page
  name or file type; e.g., 'foo.html' or '.gif', not '/foo/bar/'.
  The dir_limit and dir_exclude specifications are from the document root; 
  e.g., '/foo/bar/baz.html', not 'baz.html' and '/bat.html' not 'bat.html'.
  When setting any of these four attributes, make sure to pass the arguments
  as a [list], all at once.


- Host: client host/domain/ip/network limiting class
  hl_log = weblog.limit.Host(log object)
  methods:
    - hl_log.getlogent()
  variables:
    - hl_log.limit_host [ list of hostnames/domains to limit to ]  
    - hl_log.exclude_host [ list of hostnames/domains to exclude ]
    - hl_log.cache_size (size of page cache)
  read-only variables:
    - hl_log.num_skipped (number of lines that have been passed over)
  attributes:
    - all attributes from the log object are available.    

  Hosts can be a FQDN or domain name; e.g., 'foo.bar.com', 'bar.com', '.com'.
  Hosts are not case-sensitive.
  
  If weblog.resolve has been used, the .host attribute will be used when 
  available; otherwise, the .client attribute will be used, and will 
  not match if it is not of the correct type. 


- Time: start/end time limiting class
  tl_log = weblog.limit.Time(log object)
  methods:
	- tl_log.getlogent()
  booleans:
    - tl_log.end_stop ( if set, will getlogent() will return 0 on first
                       line matching l_log.end )
  variables:
	- tl_log.start [ exclude log lines before this unix epoch time ]
	- tl_log.end [ exclude log lines after this unix epoch time ]
  read-only variables:
    - hl_log.num_skipped (number of lines that have been passed over)
  attributes:
	- all attributes of the log object are available.

  Because Time only needs the utime attribute from the parser, it
  pays to put it at the top of the 'stack' of instances, right below
  the Parser itself.


- test: test function
'''


# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.



__version__ = '0.99'

import re
from string import join
from types import ListType


class Path:
	''' Path Limiting Class '''

	def __init__(self, log):
		self.log = log
		self.dir_limit = []
		self.dir_exclude = ['::::::::']
		self.page_limit = []
		self.page_exclude = ['::::::::']
		self.cache_size = 10000
		self.num_skipped = 0
		self._cache = {}


	def __getattr__(self, attr):
		try:
			return getattr(self.log, attr)
		except AttributeError:
			raise AttributeError, attr


	def __setattr__(self, attr, value):
		if type(value) == ListType:
			value = map(lambda a: re.escape(a), value)
		if attr == 'dir_limit':
			self._dir_limit_pat = "^(" + join(value, "|") + ")"
			self._dir_limit_comp = re.compile(self._dir_limit_pat)
		if attr == 'dir_exclude':
			self._dir_exclude_pat = "^(" + join(value, "|") + ")"
			self._dir_exclude_comp = re.compile(self._dir_exclude_pat)
		if attr == 'page_limit':
			self._page_limit_pat = "(" + join(value, "|") + ")$"
			self._page_limit_comp = re.compile(self._page_limit_pat)
		if attr == 'page_exclude':
			self._page_exclude_pat = "(" + join(value, "|") + ")$"
			self._page_exclude_comp = re.compile(self._page_exclude_pat)
		self.__dict__[attr] = value


	def getlogent(self):
		''' Increment position in the log and populate requested attributes '''

		while self.log.getlogent():
			try:
				if self._cache[self.log.url_path]:
					return 1
			except KeyError:
				if self._path_match(self.log.url_path):
					return 1
			self.num_skipped = self.num_skipped + 1
		return 0


	def _path_match(self, path):
		''' matching engine; returns 1 if page is to be ignored '''

		if len(self._cache) > self.cache_size:
			self._cache = {}
		i = 1
		if self._dir_limit_comp.search(path) == None: i = 0
		elif self._page_limit_comp.search(path) == None: i = 0
		elif self._dir_exclude_comp.search(path) != None: i = 0
		elif self._page_exclude_comp.search(path) != None: i = 0					
		self._cache[path] = i
		return self._cache[path]



class Host:
	''' Client Host/Domain Limiting Class '''
	
	def __init__(self, log):
		self.log = log
		self.host_limit = []
		self.host_exclude = ['::::::::']
		self.cache_size = 10000
		self.num_skipped = 0
		self._cache = {}
		if hasattr(self.log, 'host'):
			self._loc = 'host'
		else:
			self._loc = 'client'

		
	def __getattr__(self, attr):
		try:
			return getattr(self.log, attr)
		except AttributeError:
			raise AttributeError, attr


	def __setattr__(self, attr, value):
		if attr == 'host_limit':
			self._host_limit_pat = "(" + join(value, "|") + ")\.?$"
			self._host_limit_comp = re.compile(self._host_limit_pat, re.IGNORECASE)
		if attr == 'host_exclude':
			self._host_exclude_pat = "(" + join(value, "|") + ")\.?$"
			self._host_exclude_comp = re.compile(self._host_exclude_pat, re.IGNORECASE)
		self.__dict__[attr] = value


	def getlogent(self):
		''' Increment position in the log and populate requested attributes '''

		while self.log.getlogent():
			try:
				if self._cache[getattr(self.log, self._loc)]:
					return 1
			except KeyError:
				if self._host_match(getattr(self.log, self._loc)):
					return 1
			self.num_skipped = self.num_skipped + 1
		return 0
		

	def _host_match(self, host):
		''' matching engine; returns 1 if page is to be ignored '''

		if len(self._cache) > self.cache_size:
			self._cache = {}
		i = 1
		if self._host_limit_comp.search(host) == None: i = 0
		elif self._host_exclude_comp.search(host) != None: i = 0
		self._cache[host] = i
		return self._cache[host]
			


class Time:
	''' Start/Stop Time Limiting Class '''
	
	def __init__(self, log):
		self.log = log
		self.start = 0
		self.end = 2000000000
		self.end_stop = 0
		self.num_skipped = 0

	
	def __getattr__(self, attr):
		try:
			return getattr(self.log, attr)
		except AttributeError:
			raise AttributeError, attr


	def getlogent(self):
		''' Increment position in the log and populate requested attributes '''

		while self.log.getlogent():
			if self.log.utime < self.start:
				self.num_skipped = self.num_skipped + 1
				continue
			if self.log.utime > self.end:
				if self.end_stop:
					return 0
				else:
					self.num_skipped = self.num_skipped + 1
					continue
			return 1
		return 0	
	




def test():
	''' basic test suite- modify at will to test full functionality '''

	import sys
	from weblog import combined, url

	file = sys.stdin
	log = combined.Parser(file)

	up_log = url.Parser(log)

	lim_log = Host(up_log)
	lim_log.host_limit = ['.com']
	lim_log.host_exclude = ['aol.com']

	jpg_log = Path(lim_log)
	jpg_log.page_limit = ['.jpg']

	s_log = Time(jpg_log)
	s_log.start = 899306347

	while s_log.getlogent():
		print "%s %s" % (s_log.client, s_log.url)
	print "lines processed-", log.num_processed
	print "error lines-", log.num_error
	print "lines skipped- %s host, %s path, %s time, %s total" % (
			lim_log.num_skipped, jpg_log.num_skipped, s_log.num_skipped,
			lim_log.num_skipped + jpg_log.num_skipped + s_log.num_skipped)


if __name__ == '__main__':
	test()



#!/usr/bin/env python

'''
Squid Web proxy cache log parsing classes.

Contents:

- AccessParser: squid access logfile parser class
  log = AccessParser(log filehandle)
  - methods:
	log.getlogent()
  - attributes:
    log.utime
    log.elapsed
	log.client
	log.log_tag
	log.status
	log.bytes
	log.method
	log.url
	log.ident
	log.peer_tag
	log.peerhost
	log.mimetype
  - read-only variables:
	log.num_error
	log.num_processed

- StoreParser: squid store logfile parser class
  log = StoreParser(log filehandle)
  - methods:
	log.getlogent()
  - attributes:
    log.utime
    log.action
	log.status
	log.datehdr
	log.lastmod
	log.expires
	log.mimetype
	log.expect_len
	log.real_len
	log.method
	log.url
  - read-only variables:
	log.num_error
	log.num_processed

- test: test function
'''


# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used, 
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.


# Squid Access Logfile Format
# ---------------------------
#
# Version 1.1 Access log
# 
# timestamp elapsed_time client log_tag/status bytes method URL rfc931 \
# peer_tag/peerhost mimetype
#
# rfc931: identd info, - otherwise
#
#
# Squid Store Logfile Format
# --------------------------
#
# Version 1.1 Store log
#
# time action status datehdr lastmod expires type expect-len/real-len \
# method key
#
#
# for more information about both formats, see the Squid FAQ at
# http://squid.nlanr.net/




__version__ = '0.99'


from string import atoi, atof, split
import sys


class AccessParser:
	''' Splitting Squid Access Logfile Parser '''

	def __init__(self, file_descriptor):
		self.num_processed = 0
		self.num_error = 0
		self._fd = file_descriptor			
		self.utime = 0
		self.elapsed = 0
		self.client = ''
		self.log_tag = ''
		self.status = 0
		self.bytes = 0
		self.method = ''
		self.url = ''
		self.ident = ''
		self.peer_tag = ''
		self.peerhost = ''
		self.mimetype = ''

	
	def getlogent(self):
		''' Increament location in the log and populate object attributes '''

		while 1: 	# loop until we find a valid line, or end
			line = self._fd.readline()		
			if not line: return 0
			self.num_processed = self.num_processed + 1
			
			n = split(line, None)
			if len(n) != 10:
				self.num_error = self.num_error + 1
			else:
				try:
					self.utime = int(atof(n[0]))
					self.elapsed = int(atoi(n[1]))
					self.client = n[2]
					(self.log_tag, status) = split(n[3], '/', 2) 
					self.status = atoi(status)
					self.bytes = atoi(n[4]) 
					self.method = n[5] 
					self.url = n[6]
					self.ident = n[7]
					(self.peer_tag, self.peerhost) = split(n[8], '/', 2)
					self.mimetype = n[9]
				except:
					self.num_error = self.num_error + 1
					continue
				return 1



class StoreParser:
	''' Splitting Squid Store Logfile Parser '''

	def __init__(self, file_descriptor):
		self.num_processed = 0
		self.num_error = 0
		self._fd = file_descriptor
		self.utime = 0
		self.action = 0
		self.status = ''
		self.datehdr = ''
		self.lastmod = 0
		self.expires = 0
		self.mimetype = ''
		self.expect_len = ''
		self.real_len = ''
		self.method = ''
		self.url = ''

	
	def getlogent(self):

		''' Increament location in the log and populate object attributes '''

		while 1: 	# loop until we find a valid line, or end
			line = self._fd.readline()		
			if not line: return 0
			self.num_processed = self.num_processed + 1
			
			n = split(line, None)
			if len(n) != 10:
				self.num_error = self.num_error + 1
			else:
				try:
					self.utime = int(atof(n[0]))
					self.action = n[1]
					self.status = atoi(n[2])
					self.datehdr = atoi(n[3])
					self.lastmod = atoi(n[4]) 
					self.expires = atoi(n[5]) 
					self.mimetype = n[6]
					(expect_len, real_len) = split(n[7], '/', 2)
					self.expect_len = atoi(expect_len)
					self.real_len = atoi(real_len)
					self.method = n[8]
					self.url = n[9]
				except:
					self.num_error = self.num_error + 1
					continue
				return 1





			
def test_access():
	''' basic test suite- modify at will to test all functionality '''
	
	file = sys.stdin
	log = AccessParser(file)	
	while log.getlogent():
		print "%s %s %s %s" % (log.client, log.log_tag, log.bytes, log.peerhost)
	print "lines: %s" % (log.num_processed)
	print "error: %s" % (log.num_error)
		
		
if __name__ == '__main__':
	test_access()




#!/usr/bin/env python


'''
ConfReader - configuration file reading class

This class offers a flexible configuration file reading interface. By setting
desired parameters for configuration in one place, it helps reduce confusion
and eases development of end-user applications. It can also be used to parse
existing config files of applications, to build utilities for them.

Initialise with one argument- the name or location of the file. If it is
not absolute, the program's directory will be searched for it. In either
case, IOError will be raised if it cannot be read.

Parameters of config items can be specified by the set() function; the first
argument is the name of the config (as in the file, a string), followed by
four optional arguments:

* default- the value returned if the config is not found in the file.
* post- postprocessing function to use; can be lambda form or any function.
* required- set to 1 if config is required.
* list- set to 1 if config should always be returned as list. Multiple entries
  will be appended to the list if this is set; otherwise, the last entry will
  be used.

Although the name of the config should be specified as in the file, the 
attribute it is made available as will have any non-alphanumeric characters
replaced by an underscore ('_').

When referencing a function for postprocessing, just pass the name, without 
the parenthesis().

Defaults are set before postprocessing occurs; therefore, set defaults as
they would be seen in the file.

Once parameters have been set, call the parse() method, which will apply
them to the configs and make them available as attributes. Users of the 
parse() method should be ready to catch:

* NameError- if a required config is missing
* ValueError- if int() or float() fails
* any other postprocessing exception that may be raised

Remember that unless a default is set, None is passed to the postprocessing
function.

For postprocessing errors, the .config_line_number attribute stores the last
line of the file seen; this makes it possible to see where the error is
in the file.

Config file format:
- blank lines are ignored
- lines that start with a comment character are ignored
- comment character is # by default; set with config.comment attribute
- format of a declaration: name [separator] value
- space around the separator and at the start of the line is ignored
- separator can be changed with the config_separator attribute
- if a requested item is not declared, None will be returned
- this value can be changed with the no_value attribute

yesno() function:
yesno() can be used as a convenient postprocessing function to determine
binary values. See the function for exact behavior. It will raise a
ValueError if it cannot determine the truth of its input.


Example

>>> try:
>>> 	config = ConfReader(filename)
>>> except IOError:							# can't read file
>>> 	print "can't read", filename
>>> config.set('one_conf_item', default='foo')		# default item
>>> config.set('another_item', post=int)			# return an integer
>>> import string
>>> config.set('yet_another', post=string.lower)	# return a lower string
>>> config.set('a_nother', post=string.split)		# return a list of words
>>> config.set('and_another', post=lambda a: a)		# custom postprocessing
>>> config.set('heres_one', list=1)					# return a list
>>> config.set('last_one', required=1)				# retquired config
>>> try:
>>> 	config.parse()								# parse the file
>>> except NameError, missing_one:
>>> 	print filename, "doesn't have", missing_one	# required config missing
>>> except ValueError, convert_problem:
>>> 	print "conversion error:", convert_problem	# conversion problem
>>> print config.one_conf_item						# access a config
'foo'
>>> print config.heres_one
['value', 'value', 'value']



TODO
* Apache-style <Virtual> sections (objects)?
* end-of-line comments
* alternate input methods (any object with a readline(), from array...)
* test() function
* security (take out leading underscores?)

'''


# (c) 1998 Copyright Mark Nottingham
# <mnot@pobox.com>
#
# This software may be freely distributed, modified and used,
# provided that this copyright notice remain intact.
#
# This software is provided 'as is' without warranty of any kind.



__version__ = "0.55"

from types import *
from os import path
import sys, re

illegal_chars = re.compile('\W')	# pattern of illegal characters in
									# attribute names


class ConfReader:
	''' configuration file reading class '''

	def __init__(self, file):
		''' slurp in the conf file '''

		self.config_separator = '='
		self.config_comment = '#'
		self.config_no_value = None
		self.config_line_number = 0

		if not path.isabs(file):
			prog_name = sys.argv[0]
			prog_path = path.split(prog_name)[0]
			file = path.join(prog_path, file)
		conf = open(file, 'r')
		self._confs = conf.readlines()
		conf.close()
		self._defaults = {}
		self._post_proc = {}
		self._required = {}
		self._return_list = {}
		
	
	def __getattr__(self, attr):
		return self.config_no_value


	def set(self, name, default=None, post=None, required=0, list=0):
		''' set parameters for a configuration item '''

		self._defaults[name] = default
		self._post_proc[name] = post
		self._required[name] = required
		self._return_list[name] = list


	def parse(self):
		''' Parse the stored configuration '''

		from string import split, strip
		from types import *
		import re

		tmp = {}							# temporary config dictionary
		comment_pat = re.compile('^\s*%s' % (self.config_comment))	# comment pattern
		blank_pat = re.compile('^\s*$')		# blank lines
		self.config_line_number = 0			# reset line counter

		### process line by line
		for line in self._confs:
			if not line:
				break
			self.config_line_number = self.config_line_number + 1

			### ignore comments, blank lines
			if comment_pat.search(line):
				continue			
			if blank_pat.match(line):
				continue

			### split conf name/value	
			try:
				(conf_name, conf_value) = split(line, self.config_separator, 1)
			except ValueError:
				continue
			conf_name = strip(conf_name)
			conf_value = strip(conf_value)
			
			### add to temp dictionary
			if tmp.has_key(conf_name):
				tmp[conf_name].append(conf_value)
			else:
				tmp[conf_name] = [conf_value]

		### run through required, defaults, postprocess, list
		for name, required in self._required.items():		# required
			if not tmp.has_key(name) and required:
				raise NameError, '%s required - not found' % (name)
		for name, value in self._defaults.items():			# defaults
			if not tmp.has_key(name):
				tmp[name] = [value]
		for name, value in tmp.items():
			tr_func = self._post_proc.get(name, None)		# postprocess
			if tr_func is None: tr_func = lambda a:a
			tmp[name] = map(tr_func, value)
		for name, value in tmp.items():
			if self._return_list.get(name, 0):				# list
				if type(value) is not ListType:
					value = [value]
			else:
				if type(value) is ListType:
					value = value[-1]

			### make names legal
			name = re.sub(illegal_chars, '_', name)

			### populate attributes
			setattr(self, name, value)



def yesno(element):
	''' return 1 for true, 0 for false, or raise error '''

	import string
	e = string.lower(string.strip(element))	
	if e[0] == 't' or e == '1' or e[0] == 'y' or e == 'on':
		return 1
	elif e[0] == 'f' or e == '0' or e[0] == 'n' or e == 'off':
		return 0
	else:
		raise ValueError, "can't determine value of %s" % (element)
		

def test():
	''' haven't gotten around to this yet '''
	pass



if __name__ == '__main__':
	test()
